package org.xins.common.xml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xins.common.MandatoryArgumentChecker;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Converter from Element to String and vice versa.
 *
 * @author <a href="mailto:anthony.goubard@japplis.com">Anthony Goubard</a>
 *
 * @since XINS 3.0
 */
public class ElementFormatter {

   private ElementFormatter() {
   }

   /**
    * Returns the String representation of the element as XML.
    * The XML declaration is not included.
    *
    * @param element
    *    the element to serialize, cannot be <code>null</code>.
    *
    * @return
    *    the XML representation of the XML or an empty element if an error occured.
    */
   public static String format(Element element) {
      MandatoryArgumentChecker.check("element", element);
      try {
         TransformerFactory transformerFactory = TransformerFactory.newInstance();
         Transformer transformer = transformerFactory.newTransformer();
         transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
         transformer.setOutputProperty(OutputKeys.METHOD, "xml");
         StringWriter serializer = new StringWriter();
         Source source = new DOMSource(element.getOwnerDocument());
         transformer.transform(source, new StreamResult(serializer));
         String elementAsString = serializer.toString();
         return elementAsString;
      } catch (TransformerException ex) {
         // TODO Log
         return "<" + element.getLocalName() + "/>";
      }
   }

   /**
    * Creates a DOM element by parsing the given XML.
    *
    * @param xml
    *    the XML text to parse, cannot be <code>null</code>.
    * @return
    *    the DOM element representing the parsed XML, never <code>null</code>.
    * @throws SAXException
    *    if for any reason the XML cannot be parsed correctly.
    */
   public static Element parse(String xml) throws SAXException {
      return parse(new StringReader(xml));
   }

   /**
    * Creates a DOM element by parsing the given XML location.
    *
    * @param uri
    *    the URI location of the XML resource to parse, cannot be <code>null</code>.
    * @return
    *    the DOM element representing the parsed XML, never <code>null</code>.
    * @throws SAXException
    *    if for any reason the XML cannot be parsed correctly.
    */
   public static Element parseURI(String uri) throws SAXException {
      try {
         DocumentBuilder builder = DocumentBuilderPool.getInstance().getBuilder();
         Document document = builder.parse(uri);
         Element parsedElement = document.getDocumentElement();
         DocumentBuilderPool.getInstance().releaseBuilder(builder);
         return parsedElement;
      } catch (IOException ex) {
         throw new SAXParseException("IO error occured while parsing " + uri + ": " + ex.getMessage(), null);
      }
   }

   /**
    * Creates a DOM element by parsing the XML included in the reader.
    *
    * @param reader
    *    the reader to use to read the XML source, cannot be <code>null</code>.
    * @return
    *    the DOM element representing the parsed XML, never <code>null</code>.
    * @throws SAXException
    *    if for any reason the XML cannot be parsed correctly.
    */
   public static Element parse(Reader reader) throws SAXException {
      try {
         DocumentBuilder builder = DocumentBuilderPool.getInstance().getBuilder();
         InputSource source = new InputSource();
         source.setCharacterStream(reader);
         Document document = builder.parse(source);
         Element parsedElement = document.getDocumentElement();
         DocumentBuilderPool.getInstance().releaseBuilder(builder);
         return parsedElement;
      } catch (IOException ex) {
         throw new SAXParseException("IO error occured while parsing: " + ex.getMessage(), null);
      }
   }


   /**
    * Creates a DOM element by parsing the XML included in the input stream.
    *
    * @param stream
    *    the input stream to use to read the XML source, cannot be <code>null</code>.
    * @return
    *    the DOM element representing the parsed XML, never <code>null</code>.
    * @throws SAXException
    *    if for any reason the XML cannot be parsed correctly.
    */
   public static Element parse(InputStream stream) throws SAXException {
      try {
         DocumentBuilder builder = DocumentBuilderPool.getInstance().getBuilder();
         Document document = builder.parse(stream);
         Element rootElement = document.getDocumentElement();
         DocumentBuilderPool.getInstance().releaseBuilder(builder);
         return rootElement;
      } catch (IOException ex) {
         throw new SAXParseException("IO error occured while parsing: " + ex.getMessage(), null);
      }
   }

   /**
    * Creates an empty DOM element with the specified name.
    *
    * @param name
    *    the local name of the XML element, cannot be <code>null</code>.
    * @return
    *    a new DOM element with the specified name, never <code>null</code>.
    */
   public static Element createMainElement(String name) {
      return createMainElementNS(null, name);
   }

   /**
    * Creates an empty DOM element with the specified name.
    *
    * @param namespaceURI
    *    the namespace URI of the XML element, can be <code>null</code>.
    * @param qulifiedName
    *    the qualified name of the XML element, cannot be <code>null</code>.
    * @return
    *    a new DOM element with the specified name, never <code>null</code>.
    */
   public static Element createMainElementNS(String namespaceURI, String qualifiedName) {
      DocumentBuilder documentBuilder = DocumentBuilderPool.getInstance().getBuilder();
      Document document = documentBuilder.newDocument();
      Element element = null;
      if (namespaceURI == null) {
         element = document.createElement(qualifiedName);
      } else {
         element = document.createElementNS(namespaceURI, qualifiedName);
      }
      document.appendChild(element);
      Element createdElement = document.getDocumentElement();
      DocumentBuilderPool.getInstance().releaseBuilder(documentBuilder);
      return createdElement;
   }

   /**
    * Creates a DocumentBuilder.
    *
    * @return
    *    a newly created DocumentBuilder.
    * @deprecated use the DocumentBuilderPool
    */
   private static DocumentBuilder createDocumentBuilder() {
      // TODO Keep it in a ThreadLocal to avoid unneeded recreation?
      try {
         DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
         builderFactory.setIgnoringElementContentWhitespace(true);
         builderFactory.setNamespaceAware(true);
         builderFactory.setValidating(false);
         DocumentBuilder builder = builderFactory.newDocumentBuilder();
         builder.setEntityResolver(new EntityResolver() {
            public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException {
                return new InputSource(new ByteArrayInputStream(new byte[0]));
            }
         });
         return builder;
      } catch (ParserConfigurationException ex) {
         // Without Dom the program should fail
         // TODO logging
         throw new IllegalStateException(("Failed to create a DOM document builder: " + ex.getMessage()));
      }
   }
}
